#include <xil_exception.h>
#include <xil_io.h>
#include <xrtcpsu.h>		/* RTCPSU device driver */
#include <xstatus.h>
#include <libeswpm/test_aux.h>
#include <xtime_l.h>
#include "pm_api_sys.h"
#include "pm_client.h"
#include "time.h"
#include "sleep.h"

extern void __attribute__((weak)) *_vector_table;
static XScuGic GicInst;
static XIpiPsu IpiInst;
static XRtcPsu RtcPsu;          /* Instance of the RTC Device */
XTime tNotify;

#define XIL_PASS()				xil_printf("\r\nREGRESSION TEST PASS\r\n");

#define RTC_DEVICE_ID           XPAR_XRTCPSU_0_DEVICE_ID
#define RTC_ALARM_INT_IRQ_ID    XPAR_XRTCPSU_ALARM_INTR

#define TARGET_PU		NODE_APU
#define TARGET_NODE		NODE_APU_0
#define IPI_PMU_0_TRIG_APU_MASK (0x1)
/* Define mask encoding all target masters involved in the scenario */
#define OTHER_MASTERS_MASK	IPI_PMU_0_TRIG_APU_MASK
#define TIMER_PERIOD 		(10)
#define MAX_COUNT		(3)
/* Calculate latency from counter ticks to microseconds */
#define CALCULATE_LATENCY(x)	( (x) / (COUNTS_PER_SECOND/1000000) )

static void TestNotifyCb(XPm_Notifier* const notifier);

static XPm_Notifier notifier = {
	.callback = TestNotifyCb,
	.node = NODE_FPD,
	.event = EVENT_STATE_CHANGE,
	.flags = 0,
};

void RtcInterruptHandler(XRtcPsu timer_inst, u32 event)
{
	XRtcPsu_ClearInterruptMask(&timer_inst, XRTC_INT_EN_ALRM_MASK);
}

static XStatus TimerConfigure()
{
	XStatus status = XST_SUCCESS;
	XRtcPsu_Config *Config;

	/*
	 * Initialize the RTC driver so that it's ready to use
	 * Look up the configuration in the config table, then initialize it.
	 */
	Config = XRtcPsu_LookupConfig(RTC_DEVICE_ID);
	if (NULL == Config) {
		pm_print("Invalid config\r\n");
		return XST_FAILURE;
	}

	status = XRtcPsu_CfgInitialize(&RtcPsu, Config, Config->BaseAddr);
	if (status != XST_SUCCESS) {
		pm_print("Cfg initialization failed\r\n");
		return XST_FAILURE;
	}

	XRtcPsu_SetHandler(&RtcPsu,(XRtcPsu_Handler)RtcInterruptHandler, &RtcPsu);
	status = XScuGic_Connect(&GicInst, RTC_ALARM_INT_IRQ_ID, (Xil_ExceptionHandler)XRtcPsu_InterruptHandler, &RtcPsu);
	/* Enable the interrupt for the RTC */
	XScuGic_Enable(&GicInst, RTC_ALARM_INT_IRQ_ID);

	return status;
}

static void enableTimer()
{
	u32 CurrentTime, Alarm;

	/* Enable the interrupt of the RTC device so interrupts will occur. */
	XRtcPsu_SetInterruptMask(&RtcPsu, XRTC_INT_EN_ALRM_MASK );

	CurrentTime = XRtcPsu_GetCurrentTime(&RtcPsu);
	Alarm = CurrentTime + TIMER_PERIOD;
	XRtcPsu_SetAlarm(&RtcPsu,Alarm,0);
}

/**
 * TestNotifyCb() - Application specific callback to be chained (called from
 *		xilpm when the notifier callback fires)
 * @notifier	Pointer to the registered notifier which contains info about the
 *		notification (how many times notification has been received,
 *		what is the last received oppoint value, etc)
 */
static void TestNotifyCb(XPm_Notifier* const notifier)
{
	XTime_GetTime(&tNotify);
}

static u32 request_suspend( u32 *latency, u32 *fpd_latency)
{
	XTime tStart, tEnd;
	u32 status;

	notifier.received = 0;
	XTime_GetTime(&tStart);
	status = XPm_RequestSuspend(TARGET_PU, REQUEST_ACK_NON_BLOCKING,
				    MAX_LATENCY, 0);
	IpiWaitForAck();
	XTime_GetTime(&tEnd);

	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Request Suspend %s : %d\r\n",
				SHOW_DEFINE(TARGET_NODE), status);
		return status;
	}

	*latency = CALCULATE_LATENCY(tEnd - tStart);

	pm_print("******************** RPU ON, APU suspend with FPD ON ******************\r\n");
	pm_print("(10 seconds delay)\r\n");
	sleep(10);
	pm_print("Release NODE_SATA, FPD will be off after releasing NODE_SATA\r\n ");

	XTime_GetTime(&tStart);
	status = XPm_ReleaseNode(NODE_SATA);
	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Release Node NODE_SATA : %d\r\n",
				status);
		return status;
	}

	/* Block until the notification is received */
	while (0 == notifier.received);
	notifier.received = 0;
	pm_print("******************** RPU ON, APU suspend with FPD OFF ******************\r\n");
	*fpd_latency = CALCULATE_LATENCY(tNotify - tStart);

	return status;
}

static u32 request_wakeup(u32 *latency, u32 *pu0_latency, u32 *fpd_latency)
{
	XTime tStart, tEnd, tpu0_up;
	u32 status;

	notifier.received = 0;

	pm_print("(3 seconds delay)\r\n");
	sleep(3);

	XTime_GetTime(&tStart);
	status = XPm_RequestWakeUp(TARGET_NODE, 0, 0, REQUEST_ACK_NON_BLOCKING);
	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Request wakeup %s : %d\r\n",
				SHOW_DEFINE(TARGET_NODE), status);
		return status;
	}

	/* Block until the notification is received */
	while (0 == notifier.received);

	IpiWaitForAck();

	XTime_GetTime(&tpu0_up);
	SyncWaitForReady(OTHER_MASTERS_MASK);
	XTime_GetTime(&tEnd);
	SyncClearReady(OTHER_MASTERS_MASK);

	*fpd_latency = CALCULATE_LATENCY(tNotify - tStart);
	*pu0_latency = CALCULATE_LATENCY(tpu0_up - tNotify);
	*latency = CALCULATE_LATENCY(tEnd - tNotify);

	status = XPm_RequestNode(NODE_SATA, PM_CAP_ACCESS | PM_CAP_CONTEXT, 0, REQUEST_ACK_BLOCKING);
	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Request Node NODE_SATA : %d\r\n",
				status);
		return status;
	}

	return status;
}

static u32 measure_latency(void)
{
	u32 itteration, status, latency, pu0_latency, fpd_latency;
	u32 susp_min = ~0, susp_max = 0, susp_avg = 0;
	u32 fpd_off_min = ~0, fpd_off_max = 0, fpd_off_avg = 0;
	u32 wake_min = ~0, wake_max = 0, wake_avg = 0;
	u32 pu0_wake_min = ~0, pu0_wake_max = 0, pu0_wake_avg = 0;
	u32 fpd_on_min = ~0, fpd_on_max = 0, fpd_on_avg = 0;
	u64 susp_total = 0, wake_total = 0, pu0_wake_total = 0;
	u64 fpd_on_total = 0, fpd_off_total=0;

	pm_print("measure_latency start\r\n");
	for (itteration = 0; itteration < MAX_COUNT; itteration++) {

		/* Measure suspend latency */
		status = request_suspend(&latency, &fpd_latency);
		if (status != XST_SUCCESS)
			return XST_FAILURE;

		susp_min = (latency < susp_min ) ? latency : susp_min;
		susp_max = (latency > susp_max ) ? latency : susp_max;
		susp_total += latency;

		fpd_off_min = (fpd_latency < fpd_off_min ) ? fpd_latency : fpd_off_min;
		fpd_off_max = (fpd_latency > fpd_off_max ) ? fpd_latency : fpd_off_max;
		fpd_off_total += fpd_latency;

		pm_print("Suspended.\r\n");

		pm_print("(10 seconds delay)\r\n");
		sleep(10);

		pm_print("Resuming APU.\r\n");
		/* Measure wakeup latency */
		status = request_wakeup(&latency, &pu0_latency, &fpd_latency);
		if (status != XST_SUCCESS)
			return XST_FAILURE;

		pm_print("Resumed.\r\n");
		wake_min = (latency < wake_min ) ? latency : wake_min;
		wake_max = (latency > wake_max ) ? latency : wake_max;
		wake_total += latency;

		pu0_wake_min = (pu0_latency < pu0_wake_min ) ? pu0_latency : pu0_wake_min;
		pu0_wake_max = (pu0_latency > pu0_wake_max ) ? pu0_latency : pu0_wake_max;
		pu0_wake_total += pu0_latency;

		fpd_on_min = (fpd_latency < fpd_on_min ) ? fpd_latency : fpd_on_min;
		fpd_on_max = (fpd_latency > fpd_on_max ) ? fpd_latency : fpd_on_max;
		fpd_on_total += fpd_latency;

		pm_print("(10 seconds delay)\r\n");
		sleep(10);
	}

	susp_avg = susp_total / MAX_COUNT;
	wake_avg = wake_total / MAX_COUNT;
	pu0_wake_avg = pu0_wake_total / MAX_COUNT;
	fpd_off_avg = fpd_off_total / MAX_COUNT;
	fpd_on_avg = fpd_on_total / MAX_COUNT;

	pm_print("Request Suspend Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), susp_min, susp_max, susp_avg);
	pm_print("FPD off Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), fpd_off_min, fpd_off_max, fpd_off_avg);
	pm_print("FPD on Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), fpd_on_min, fpd_on_max, fpd_on_avg);
	pm_print("APU0 Wakeup Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), pu0_wake_min, pu0_wake_max, pu0_wake_avg);
	pm_print("Wakeup Latency in useconds of Node %s:  Min: %ld, Max: %ld, Avg: %ld\r\n",
			SHOW_DEFINE(TARGET_NODE), wake_min, wake_max, wake_avg);

	return XST_SUCCESS;
}

/**
 * prepare_suspend() - request suspend
 */
static void prepare_suspend(void)
{
	XStatus status;

	status = XPm_SetWakeUpSource(NODE_RPU_0, NODE_RTC, 1);
	HALT_IF_ERR(status, "in enable wake-up source NODE_RTC");

	status = XPm_SelfSuspend(NODE_RPU_0, MAX_LATENCY, 0, (u64)&_vector_table);
	HALT_IF_ERR(status, "in self suspend call for RPU_0");

	status = XPm_SetRequirement(NODE_TCM_0_A, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_TCM_0_A");

	status = XPm_SetRequirement(NODE_TCM_0_B, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_TCM_0_B");

	status = XPm_SetRequirement(NODE_TCM_1_A, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_TCM_1_A");

	status = XPm_SetRequirement(NODE_TCM_1_B, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_TCM_1_B");

}

int main(void)
{
	u32 status;
	int ret = XST_SUCCESS;
	enum XPmBootStatus boot;

	boot = XPm_GetBootStatus();
	if (PM_INITIAL_BOOT == boot) {

		/* Initialize GIC, IPIs, and Xilpm */
		PmTestInit(&GicInst, &IpiInst);

		status = XPm_RequestNode(NODE_UART_0, PM_CAP_ACCESS | PM_CAP_CONTEXT, 0, REQUEST_ACK_BLOCKING);
		pm_print("status #%d\r\n", status);

		pm_print("INITIAL BOOT\r\n");
		pm_print("************* ZynqMP Typical Power State Demo Started ********************\r\n");

		status = XPm_RequestNode(NODE_SATA, PM_CAP_ACCESS | PM_CAP_CONTEXT, 0, REQUEST_ACK_BLOCKING);
		pm_print("status #%d\r\n", status);

		/* Release unused nodes which are requested from pre alloc section */
		status = XPm_ReleaseNode(NODE_L2);
		HALT_IF_ERR(status, "in release node for NODE_L2");
		status = XPm_ReleaseNode(NODE_DDR);
		HALT_IF_ERR(status, "in release node for NODE_DDR");
		status = XPm_ReleaseNode(NODE_PL);
		HALT_IF_ERR(status, "in release node for NODE_PL");
		status = XPm_ReleaseNode(NODE_OCM_BANK_0);
		HALT_IF_ERR(status, "in release node for NODE_OCM_BANK_0");
		status = XPm_ReleaseNode(NODE_OCM_BANK_1);
		HALT_IF_ERR(status, "in release node for NODE_OCM_BANK_1");
		status = XPm_ReleaseNode(NODE_OCM_BANK_2);
		HALT_IF_ERR(status, "in release node for NODE_OCM_BANK_2");
		status = XPm_ReleaseNode(NODE_OCM_BANK_3);
		HALT_IF_ERR(status, "in release node for NODE_OCM_BANK_3");
		status = XPm_ReleaseNode(NODE_I2C_0);
		HALT_IF_ERR(status, "in release node for NODE_I2C_0");
		status = XPm_ReleaseNode(NODE_I2C_1);
		HALT_IF_ERR(status, "in release node for NODE_I2C_1");
		status = XPm_ReleaseNode(NODE_SD_1);
		HALT_IF_ERR(status, "in release node for NODE_SD_1");
		status = XPm_ReleaseNode(NODE_QSPI);
		HALT_IF_ERR(status, "in release node for NODE_QSPI");
		status = XPm_ReleaseNode(NODE_ADMA);
		HALT_IF_ERR(status, "in release node for NODE_ADMA");

		status = XPm_RegisterNotifier(&notifier);
		HALT_IF_ERR(status, "in register notifier");


	} else if (PM_RESUME == boot) {
		status = XPm_RequestNode(NODE_UART_0, PM_CAP_ACCESS | PM_CAP_CONTEXT, 0, REQUEST_ACK_BLOCKING);
		pm_print("status #%d\r\n", status);

		pm_print("RESUMED\r\n");
		/* Timer is already counting, just enable interrupts */
		pm_print("Trigger enabling interrupts\r\n");
		Xil_ExceptionEnable();

		/* Request NODE_SATA */
		status = XPm_RequestNode(NODE_SATA, PM_CAP_ACCESS | PM_CAP_CONTEXT, 0, REQUEST_ACK_BLOCKING);
		pm_print("status #%d\r\n", status);

		pm_print("FPD is on now.\r\n");
		pm_print("(10 seconds delay)\r\n");
		sleep(10);
		pm_print("Resume APU\r\n");
		status = XPm_RequestWakeUp(TARGET_NODE, 0, 0, REQUEST_ACK_NON_BLOCKING);
		if (status != XST_SUCCESS) {
			pm_print("Incorrect status returned for Request wakeup %s : %d\r\n",
					SHOW_DEFINE(TARGET_NODE), status);
			return status;
		}
		/* Block until other masters in the system become ready */
		SyncWaitForReady(OTHER_MASTERS_MASK);
		SyncClearReady(OTHER_MASTERS_MASK);

		pm_print("**************** Demo executed successfully ********************\r\n");
		pm_print("******** Executing Typical power state demo again **************\r\n");
	} else {
		status = XST_FAILURE;
		HALT_IF_ERR(status, "cannot identify boot reason");
	}

	/* Configure timer, if configuration fails abort */
	status = TimerConfigure();
	HALT_IF_ERR(status, "setting up timer");

	/* Block until other masters in the system become ready */
	SyncWaitForReady(OTHER_MASTERS_MASK);
	SyncClearReady(OTHER_MASTERS_MASK);

	status = measure_latency();
	if (status != XST_SUCCESS) {
		pm_print("latency measurement failed for %s\r\n", SHOW_DEFINE(TARGET_PU));
		ret = status;
	}

	pm_print("Latency measurement done\r\n");

	status = XPm_ReleaseNode(NODE_SATA);
	if (status != XST_SUCCESS) {
		pm_print("Incorrect status returned for Release Node NODE_SATA : %d\r\n",
				status);
		return status;
	}

	pm_print ("Turn Off FPD\r\n");
	status = XPm_RequestSuspend(TARGET_PU, REQUEST_ACK_NON_BLOCKING,
			MAX_LATENCY, 0);
	IpiWaitForAck();

	/* Enable RTC timer */
	enableTimer();

	pm_print("******************** RPU idle, APU suspend with FPD OFF ******************\r\n");
	pm_print("Go to RPU Sleep mode \r\n");
	pm_print("(10 seconds timer)\r\n");
	__asm__("wfi");

	/* Enable RTC timer */
	enableTimer();

	pm_print("*************************** Deep Sleep Mode *****************************\r\n");
	pm_print("Go to deep sleep\r\n");
	prepare_suspend();
	pm_print("(10 seconds timer)\r\n");
	XPm_ReleaseNode(NODE_UART_0);
	XPm_SuspendFinalize();

	return ret;
}
